# Declaration order is based on the dependency on static references
#
# To Do
# ~~~~~
# - None, yet !

# from .audio_io import (
#     PE_AudioFilter, PE_AudioResampler,
#     PE_FFT, PE_GraphXY,
#     PE_MicrophonePA, PE_MicrophoneSD, PE_Speaker
# )

from .image_io import (
    ImageReadFile
)

# from .video_io import VideoReadFile, VideoShow, VideoWriteFile



# Usage
# ~~~~~
# cd src/aiko_services/elements/media
# aiko_pipeline create pipeline_image_io.json --stream_id 1  \
#   -sp ImageReadFile.path image.jpeg
#   -sp ImageResize.width 320 -sp ImageResize.height 240
#
# To Do
# ~~~~~
# - Document "DataSource" and "DataTarget" ...
#   - ImageReadFile: start_stream() --> create_frame() --> process_frame()
#                    Therefore pass parameters as required
#   - ImageWriteFile: Mustn't use create_frame()
#                     Therefore get parameters in process_frame()
#
# - ImageReadFile should accept a DataSource type ...
#   - URL: "file://" and media_type: "jpeg", "png", etc
#
# - PE_Metrics: Determine what to measure ?

from typing import Tuple
from pathlib import Path
from PIL import Image

import aiko_services as aiko
_LOGGER = aiko.get_logger(__name__)

_NUMPY_IMPORTED = False
try:
    import numpy as np
    _NUMPY_IMPORTED = True
except:  # TODO: Optional warning flag to avoid being annoying !
    diagnostic = "image_io.py: Couldn't import numpy module"
    print(f"WARNING: {diagnostic}")
    _LOGGER.warning(diagnostic)

def containsAll(source: str, match: chr):
    return 0 not in [character in source for character in match]

# --------------------------------------------------------------------------- #
# Check
# ~~~~~
# * aiko_pipeline create -s 1 -sp path --> load the stream parameter data
#   - Overriding the PipelineDefinition
#   - Same for HL Serving Tasks and HL Live
#
# *#1 PE.set_parameter(): parameters --> self.share[] ?
#
# * Consider process_frame() updating stream parameters --> self.share[]
#
# * For "PipelineElement.create_frames()" ...
#   - Implement frame window, which has a maximum size to minimise frames
#   - Implement frame "push back" so subsequent PEs can manage frame flow
#
# * If DataSource has "stream_required" ...
#   - Each DataSource developer shouldn't have to do this
#
# - Try to remove the need to write PE.__init__(...) in every PE
#
# - Check HL Serving Task and HL Live
#   - process_datasources(stream_id, parameters, [data_sources])
#     - data_sources: URL and MediaType
#
# - aiko_pipeline argument options for AIKO_LOG_LEVEL / AIKO_LOG_MQTT
#   - AIKO_LOG_MQTT --> AIKO_LOG=mqtt,console,...
#
# - Return Pipeline result back to HL Serving Task
#
# - Read / Write, to/from File, to/from Stream
#
# - URLs
#   - file://, mqtt://, hl:// (pre-signed URL), s3://
#   - web_camera://, rtsp://, webrt://
#   - youtube:// uploaded video or live stream
#   * Haplomic: Shared memory, e.g shm://
#
# - MediaTypes
#   - image/raw|jpeg|png, video/h.264, audio/*, text/raw, pointcloud/2d|3d
#   * Laboratory camera: image/raw

# --------------------------------------------------------------------------- #
# ImageReadFile is a DataSource with an image "path" string parameter
#
# Supports both Streams and direct process_frame() calls
#
# To Do
# ~~~~~
# - Add Stream "parameter" for optional output Image numpy format
# - Consider what causes Stream to be closed, e.g single frame processed ?
#
# Test
# ~~~~
# export AIKO_LOG_LEVEL=DEBUG; export AIKO_LOG_MQTT=false
# aiko_pipeline create pipeline_image_io_0.json -fd "(path: z_data/in_00.jpeg)"
# aiko_pipeline create pipeline_image_io_0.json -s 1  \
#                                  -sp ImageReadFile.path z_data/in_00.jpeg

class ImageReadFile(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.get_implementation("PipelineElement").__init__(self, context)

    def start_stream(self, stream, stream_id):
        path, found = self.get_parameter("path")
        if not found:
            diagnostic = 'Must provide image "path" parameter'
            return aiko.StreamEvent.ERROR, diagnostic

        self.create_frame(stream, {"path": path})
        return aiko.StreamEvent.OKAY, None

    def process_frame(self, stream, path) -> Tuple[StreamEvent, dict]:
        self.logger.debug(f"{self.my_id()}: image path: {path}")

        if not Path(path).exists():
            diagnostic = f'Image "{path}" does not exist'
            return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}
        try:
            image = Image.open(path)
        except Exception as exception:
            diagnostic = f"Error loading image: {exception}"
            return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

        self.logger.info(f"Image shape: {image.size}")
        return aiko.StreamEvent.OKAY, {"image": image}

# --------------------------------------------------------------------------- #
# ToDo: Display lines, mask, polygons, rectangles, text
# ToDo: Display Metrics
# ToDo: Display pose stick figures

class ImageOverlay(aiko.PipelineElement):
    pass

# --------------------------------------------------------------------------- #
# ToDo: Add logic for using different backends (opencv,...) or
#           different input types (np.ndarray, ...)
# ToDo: cv2.resize(image, dimensions, interpolation=cv2.INTER_CUBIC) ?

class ImageResize(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, image) -> Tuple[StreamEvent, dict]:
        width, _ = self.get_parameter("width")
        height, _ = self.get_parameter("height")

        image = image.resize((int(width), int(height)))
        return aiko.StreamEvent.OKAY, {"image": image}

# --------------------------------------------------------------------------- #
# ImageWriteFile is a DataTarget that writes an image to a file
#
# in:  "image" to be written to a file
# out: None
#
# parameter: "path" is the write file path, format variable: "frame_id"
#
# Supports both Streams and direct process_frame() calls
#
# Test
# ~~~~
# export AIKO_LOG_LEVEL=DEBUG; export AIKO_LOG_MQTT=false
# aiko_pipeline create pipeline_image_io_0.json -s 1  \
#                                  -sp ImageWriteFile.path z_data/out_01.jpeg

class ImageWriteFile(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, image) -> Tuple[StreamEvent, dict]:
        path, found = self.get_parameter("path")
        if not found:
            diagnostic = 'Must provide image write file "path" parameter'
            return aiko.StreamEvent.ERROR, diagnostic

        if containsAll(path, "{}"):
            path = path.format(stream["frame_id"])
        self.logger.debug(f"{self.my_id()}: image write file path: {path}")

        if not isinstance(image, Image.Image):
            if isinstance(image, np.ndarray):  # TODO: Check NUMPY_IMPORTED
                pass                           # TODO: numpy conversion
            else:
                diagnostic = "UNKNOWN IMAGE TYPE"  # FIX ME !
                return aiko.StreamEvent.ERROR, diagnostic

        try:
            image.save(path)
        except Exception as exception:
            diagnostic = f"Error saving image: {exception}"
            return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

        self.logger.info(f"Image shape: {image.size}")
        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #



# Definition(s)
# ~~~~~~~~~~~~~
# swag: One's personal bundle of belongings (Australian lingo)
# SWAG: Stuff We All Get
# SWAG: https://en.wikipedia.org/wiki/Scientific_wild-ass_guess :)
#
# Notes
# ~~~~~
# - Watch out for "stream.py:Frame()" dataclass and
#   "dashboard.py" use of "asciimatics.widgets.Frame()"
#
# To Do
# ~~~~~
# - Refactor from "pipeline.py", extract Stream concepts including Parameters
#   - Review "../archive/main/stream_2020.py"

from dataclasses import dataclass, field
from typing import Any, Dict

__all__ = [
    "DEFAULT_STREAM_ID", "FIRST_FRAME_ID", "Frame", "Stream",
    "StreamEvent", "StreamEventName", "StreamState", "StreamStateName"
]

DEFAULT_STREAM_ID = "*"  # string (or bytes ?)
FIRST_FRAME_ID = 0       # integer

class StreamEvent:
    ERROR =   -2  # Move to StreamState.ERROR
    STOP  =   -1  # Move to StreamState.STOP
    OKAY  =    0  # Stay calm and keep on running
    USER  = 1000  # User defined custom StreamEvents start from here

StreamEventName = {
    StreamEvent.ERROR: "Error",
    StreamEvent.OKAY:  "Okay",
    StreamEvent.STOP:  "Stop",
    StreamEvent.USER:  "User"
}

class StreamState:
    ERROR =   -2  # Don't generate new frames and ignore queued frames
    STOP  =   -1  # Don't generate new frames and process queued frames
    RUN   =    0  # Generate new frames and process queued frames
    USER  = 1000  # User defined custom StreamStates start from here

StreamStateName = {
    StreamState.ERROR: "Error",
    StreamState.STOP:  "Stop",
    StreamState.RUN:   "Run",
    StreamState.USER:  "User"
}

@dataclass
class Frame:  # effectively a continuation :)
    frame_id: int = FIRST_FRAME_ID
    parameters: Dict[str, Any] = field(default_factory=dict)
    paused_pe_name: str = None
    metrics: Dict[str, Any] = field(default_factory=dict)
    swag: Dict[str, Any] = field(default_factory=dict)
    variables: Dict[str, Any] = field(default_factory=dict)

@dataclass
class Stream:
    stream_id: str = DEFAULT_STREAM_ID
    frame_id: int = FIRST_FRAME_ID  # main thread only
    frames: Dict[int, Frame] = field(default_factory=dict)
    parameters: Dict[str, Any] = field(default_factory=dict)  # initial
    state: StreamState = StreamState.RUN
#   topic_response



# Usage
# ~~~~~
# cd aiko_services/elements
# aiko_pipeline create pipeline_text_io_0.json --stream_id 1  \
#   --stream_parameters TextReadFile.path text.txt
#
# To Do
# ~~~~~
# - TextReadFile(s): Single file or list of files or directory
#   - Option: Each line is a record (streaming)
#   - Formatted as CR/LF records, JSON, XML, CSV
# - TextWriteFile(s): Single file or list of files or directory
#   - Option: Each line is a record (streaming)
#   - Formatted as CR/LF records, JSON, XML, CSV
# - TextFilters: strip(), lower(), upper(), line/word/character count, ...
# - Pre-processing / Post-processing, e.g abbrevations, acronyms
#   - Speech-To-Text: Words not easily recognised
#     - Example: Microphone --> Speech-To-Text --> Text
#   - Text-To-Speech: Poor pronouciation
#     - Example: Text --> Text-To-Speech --> Speaker
# - Text Framing for LLM

from typing import Tuple
from pathlib import Path

import aiko_services as aiko

# --------------------------------------------------------------------------- #
# TextReadFile is a DataSource with a text file "path" string parameter
#
# To Do
# ~~~~~
# - Check: Supports both Streams and direct process_frame() calls
#
# Test
# ~~~~
# export AIKO_LOG_LEVEL=DEBUG; export AIKO_LOG_MQTT=false
# aiko_pipeline create pipeline_text_io.json -fd "(path: z_in_01.txt)"
# aiko_pipeline create pipeline_text_io.json -s 1
# aiko_pipeline create pipeline_text_io.json -s 1  \
#                                         -sp TextReadFile.path z_in_01.txt

class TextReadFile(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.get_implementation("PipelineElement").__init__(self, context)

    def start_stream(self, stream, stream_id):
        path, found = self.get_parameter("path")
        if not found:
            diagnostic = 'Must provide text file "path" parameter'
            return aiko.StreamEvent.ERROR, diagnostic

        self.create_frame(stream, {"path": path})
        return aiko.StreamEvent.OKAY, None

    def process_frame(self, stream, path) -> Tuple[aiko.StreamEvent, dict]:
        self.logger.debug(f"{self.my_id()}: text file: {path}")

        if not Path(path).exists():
            diagnostic = f'Text file "{path}" does not exist'
            return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}
        try:
            with open(path, "r") as file:
                text = file.read()
        except Exception as exception:
            diagnostic = f"Error loading text file: {exception}"
            return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

        self.logger.debug(f"Text: {text}")
        return aiko.StreamEvent.OKAY, {"text": text}

# --------------------------------------------------------------------------- #

class TextFilter(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, text) -> Tuple[aiko.StreamEvent, dict]:
    #   option, _ = self.get_parameter("option")
        text = text.upper()
        return aiko.StreamEvent.OKAY, {"text": text}

# --------------------------------------------------------------------------- #
# TextWriteFile is a DataTarget with a text string parameter
#
# To Do
# ~~~~~
# - Check: Supports both Streams and direct process_frame() calls
#
# - File path template, e.g variable frame_id
# - Consider what causes Stream to be closed, e.g single frame processed ?
#
# Test
# ~~~~
# export AIKO_LOG_LEVEL=DEBUG; export AIKO_LOG_MQTT=false
#                                          -sp TextWriteFile.path z_out_01.txt

class TextWriteFile(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, text) -> Tuple[aiko.StreamEvent, dict]:
        path, found = self.get_parameter("path")
        if not found:
            diagnostic = 'Must provide text file "path" parameter'
            return aiko.StreamEvent.ERROR, diagnostic

    #   path = path_template  # TODO: Implement path_template processing

        self.logger.debug(f"{self.my_id()}: text file path: {path}")

        try:
            with open(path, "w") as file:
                file.write(text)
        except Exception as exception:
            diagnostic = f"Error saving text file: {exception}"
            return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

        self.logger.debug(f"Text: {text}")
        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #



# To Do
# ~~~~~
# - Update to latest Pipeline API
#
# - Implement ...
#     video_capture = cv2.VideoCapture(video_pathname)
#     width = int(video_capture.get(cv2.CAP_PROP_FRAME_WIDTH))
#     height = int(video_capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
#     length = int(video_capture.get(cv2.CAP_PROP_FRAME_COUNT))
#     frame_rate = int(video_capture.get(cv2.CAP_PROP_FPS))
#
# - Design video windowing, i.e collecting multiple frame for ML processing
#   together, e.g gesture analysis
#
# - "Batching" images for CPU-GPU memory transfer efficiently for nVidia
#   - For both images and video ... depending upon ML model
#
# Framing ?

try:
    import cv2
except ModuleNotFoundError:
    raise ModuleNotFoundError((
        "opencv-python package not installed. "
        'Install aiko_services with --extras "opencv" '
        "or install opencv-python manually to use this "
        "module."))

from pathlib import Path

import aiko_services as aiko

__all__ = ["VideoReadFile", "VideoShow", "VideoWriteFile"]

# class VideoSample():
# Implement as a function which can be called from VideoRead* DataSources

# TODO: Add sample rate parameter
class VideoReadFile(StreamElement):
    def stream_start_handler(self, stream_id, frame_id, swag):
        self.logger.debug(f"stream_start_handler(): stream_id: {stream_id}")
        video_pathname = self.parameters["video_pathname"]
        self.video_capture = cv2.VideoCapture(video_pathname)
        if (self.video_capture.isOpened() == False):
            self.logger.error(f"Couldn't open video file: {video_pathname}")
            return False, None

        self.state_action = None
        if self.pipeline_state_machine and "state_action" in self.parameters:
            self.state_action = self.parameters["state_action"]
        return True, None

    def stream_frame_handler(self, stream_id, frame_id, swag):
        if self.video_capture.isOpened():
            success, image_bgr = self.video_capture.read()
            if success == True:
                self.logger.debug(f"stream_frame_handler(): stream_id: {stream_id}, frame_id: {frame_id}")
                image_rgb = cv2.cvtColor(image_bgr, cv2.COLOR_BGR2RGB)
                if frame_id % 10 == 0:
                    print(f"Frame Id: {frame_id}", end="\r")

                if self.state_action:
                    if frame_id == self.state_action[0]:
                        self.pipeline_state_machine.transition(self.state_action[1], None)
                return True, {"image": image_rgb}
            else:
                self.logger.debug("End of video")
        return False, None

    def stream_stop_handler(self, stream_id, frame_id, swag):
        self.logger.debug(f"stream_stop_handler(): stream_id: {stream_id}")
        self.video_capture.release()
        self.video_capture = None
        return True, None

class VideoShow(StreamElement):
    def stream_frame_handler(self, stream_id, frame_id, swag):
        self.logger.debug(f"stream_frame_handler(): stream_id: {stream_id}, frame_id: {frame_id}")
        title = self.parameters["window_title"]
        image_rgb = swag[self.predecessor]["image"]
        image_bgr = cv2.cvtColor(image_rgb, cv2.COLOR_BGR2RGB)
        cv2.imshow(title, image_bgr)
        if frame_id == 0:
            window_x = self.parameters["window_location"][0]
            window_y = self.parameters["window_location"][1]
            cv2.moveWindow(title, window_x, window_y)
        if cv2.waitKey(1) & 0xFF == ord("q"):
            return False, None
        return True, {"image": image_rgb}

    def stream_stop_handler(self, stream_id, frame_id, swag):
        self.logger.debug(f"stream_stop_handler(): stream_id: {stream_id}")
        cv2.destroyAllWindows()
        return True, None

class VideoWriteFile(StreamElement):
    def stream_start_handler(self, stream_id, frame_id, swag):
        self.logger.debug(f"stream_start_handler(): stream_id: {stream_id}")
        self.image_shape = None
        self.video_format = self.parameters.get("video_format", "AVC1")
        self.video_frame_rate = self.parameters["video_frame_rate"]
        self.video_pathname = self.parameters["video_pathname"]
        self.video_writer = None
        return True, None

    def _init_video_writer(self, video_pathname, video_format, frame_rate, image_shape):
        video_directory = Path(video_pathname).parent
        video_directory.mkdir(exist_ok=True, parents=True)
        return cv2.VideoWriter(
                video_pathname,
                cv2.VideoWriter_fourcc(*video_format),
                frame_rate,
                image_shape)

    def stream_frame_handler(self, stream_id, frame_id, swag):
        self.logger.debug(f"stream_frame_handler(): stream_id: {stream_id}, frame_id: {frame_id}")
        image_rgb = swag[self.predecessor]["image"]
        image_bgr = cv2.cvtColor(image_rgb, cv2.COLOR_BGR2RGB)

        if self.video_writer is None:
            if self.image_shape is None:
                self.image_shape = (image_rgb.shape[1], image_rgb.shape[0])
            self.video_writer = self._init_video_writer(
                    self.video_pathname,
                    self.video_format,
                    self.video_frame_rate,
                    self.image_shape)

        self.video_writer.write(image_bgr)
        return True, {"image": image_rgb}

    def stream_stop_handler(self, stream_id, frame_id, swag):
        self.logger.debug(f"stream_stop_handler(): stream_id: {stream_id}")
        if self.video_writer:
            self.video_writer.release()
            self.video_writer = None
        return True, None
